#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Deployments client module."""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.session import Session


class Deployment(BaseClient):
    """Client class for interacting with Deployments."""

    BASE_ENDPOINT = "v1/deployments"

    @overload
    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        kind: Optional[str] = None,
        term: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        kind: Optional[str] = None,
        term: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        kind: Optional[str] = None,
        term: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """List all App & Block Deployments ordered by creation date. Returns info on the latest revision of each deployment.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/Deployments/get_v1_deployments>`__

        Args:
            status (List[str] | None): List of statuses to filter deployments.
                Valid values are `deploying`, `deployed`, `deleting`, `delete_failed`, `failed`, `platform_resource_error`, `redeploying`, `rollback`, `rollback_complete`, `rollback_failed`, and `warning`.
            kind (str | None): Return deployments of the kind specified.
            term (str | None): Return deployments containing the specified term in name, title, description and summary.
            sort (List[str] | None): List of fields with desired ordering in the format `[<field>:<order>, ...]`,
                where `order` is one of `['asc', 'desc']` and field is an ordered parameter within the response,
                defaults to `[]`. Valid fields are `createdAt`, `createdBy`, `name` and `title`.
            page_size (int | None): Number of deployments to include per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of deployments for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: An iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
            StopIteration: There are no more pages to list.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/"
        params: Dict[str, Any] = {
            "status": status,
            "kind": kind,
            "term": term,
            "sort": sort,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="deployments",
                params=params,
                subdomain="press",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
            subdomain="press",
        )

    def execute_resources(self, deployment_id: str) -> Dict[str, Any]:
        """Trigger the resources of an app or block deployment. This will execute the resources of the latest revision of the deployment.

        Args:
            deployment_id (str): ID of the app or block deployment to trigger resources for.

        Returns:
            Dict[str, Any]: A dictionary containing the details of the triggered deployment.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/{deployment_id}/execute"
        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )


def get_client(session: Optional[Session] = None) -> Deployment:
    """Returns a Deployment client.

    Args:
        session (Optional[Session]): A Session Object. If no session is provided, a default session is used.

    Returns:
        Deployment: the Deployment client object
    """
    return Deployment(session)


__all__: List[str] = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Peak specs commands."""
from typing import List, Optional

import typer
from peak.cli import args
from peak.cli.args import OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes
from peak.output import Writer
from peak.press.specs import Spec
from rich.console import Console

app = typer.Typer(
    help="Manage both Block and App specs.",
    short_help="Manage both Block and App specs.",
)
console = Console()

RELEASE_VERSION = typer.Option(None, help="The release version of the spec in valid semantic versioning format.")


@app.command("list", short_help="List App and Block specs.")
def list_specs(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    status: Optional[List[str]] = args.STATUS_FILTER_SPECS,
    kind: Optional[str] = args.KIND_FILTER,
    term: Optional[str] = args.TERM_FILTER,
    sort: Optional[List[str]] = args.SORT_KEYS,
    scope: Optional[List[str]] = args.SCOPES,
    featured: Optional[bool] = args.FEATURED,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all the App and Block specs that exists for the tenant.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak specs list --featured --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "specCount": 1,
        "specs": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 10
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Specs/get_v1_specs)
    """
    spec_client: Spec = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = spec_client.list_specs(
            status=status,
            featured=featured,
            kind=kind,
            term=term,
            sort=sort,
            scope=scope,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="List deployments for a spec release.")
def list_release_deployments(
    ctx: typer.Context,
    spec_id: str = args.SPEC_ID,
    version: Optional[str] = RELEASE_VERSION,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    status: Optional[List[str]] = args.STATUS_FILTER_SPEC_RELEASES,
    name: Optional[str] = args.NAME_FILTER,
    title: Optional[str] = args.TITLE_FILTER,
    sort: Optional[List[str]] = args.SORT_KEYS,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all the deployments for a given spec. Version is optional and if not provided, all deployments for the spec will be returned.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak specs list-release-deployments --spec-id "632a4e7c-ab86-4ecb-8f34-99b5da531ceb" --status deployed,failed --version 1.0.0
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "deploymentCount": 1,
        "deployments": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 10,
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Specs/get_v1_specs__specId__releases__version__deployments)
    """
    spec_client: Spec = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = spec_client.list_spec_release_deployments(
            spec_id=spec_id,
            version=version,
            status=status,
            name=name,
            title=title,
            sort=sort,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Session module for Peak API."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional

from peak import exceptions
from peak.constants import DOWNLOAD_CHUNK_SIZE, ContentType, HttpMethods, Stage
from peak.handler import Handler
from peak.helpers import get_base_domain
from peak.logger import logger

DEFAULT_SESSION = None


def _get_default_session() -> Session:
    """Get the global default session object. Creates one if not already created and re-uses it.

    Returns:
        Session: The default session object
    """
    global DEFAULT_SESSION  # noqa: PLW0603
    if DEFAULT_SESSION is None:
        logger.debug("Creating DEFAULT_SESSION object")
        DEFAULT_SESSION = Session()

    logger.debug("DEFAULT_SESSION already present, reusing the object")
    return DEFAULT_SESSION


class Session:
    """A session stores credentials which are used to authenticate the requests.

    By default, a DEFAULT_SESSION is created which reads the credentials from the env variables.
    Custom Session objects can be created and used if you want to work with multiple tenants.
    """

    auth_token: str
    stage: Stage
    base_domain: str
    handler: Handler

    def __init__(
        self,
        auth_token: Optional[str] = None,
        stage: Optional[str] = None,
    ) -> None:
        """Initialize a session for the Peak API.

        Args:
            auth_token (str | None): Authentication token. Both API Key and Bearer tokens are supported.
                Picks up from `PEAK_AUTH_TOKEN` or `API_KEY` environment variable if not provided.
            stage (str | None): Name of the stage where tenant is created. Default is `prod`.
        """
        self.base_domain: str = ""
        self._set_auth_token(auth_token)
        self._set_stage(stage)
        self._set_base_domain()
        self.handler = Handler()

    def create_request(
        self,
        endpoint: str,
        method: HttpMethods,
        content_type: ContentType,
        *,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
        path: Optional[str] = None,
        ignore_files: Optional[list[str]] = None,
        subdomain: Optional[str] = "service",
    ) -> Any:
        """Prepares a request to be sent over the network.

        Adds auth_token to the headers and creates URL using STAGE.
        To be used with endpoints which returns JSON parsable response.

        Args:
            endpoint (str): The endpoint to send the request to.
            method (HttpMethods): The HTTP method to use.
            content_type (ContentType): The content type of the request.
            params (Dict[str, Any], optional): params to send to the request, defaults to None
            body (Dict[str, Any], optional): body to send to the request, defaults to None
            path (Optional[str] optional): path to the file or folder that will be compressed and used as artifact, required for multipart requests.
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact, used only for multipart requests.
            subdomain (Optional[str]): Subdomain for the endpoint. Defaults to `service`.

        Returns:
            Any: response dict object.
        """
        headers: Dict[str, str] = {"Authorization": self.auth_token}
        base_domain: str = get_base_domain(stage=self.stage.value, subdomain=subdomain)
        url: str = f"{base_domain}/{endpoint}"
        return self.handler.make_request(
            url,
            method,
            content_type=content_type,
            headers=headers,
            params=params or {},
            body=body or {},
            path=path,
            ignore_files=ignore_files,
            session_meta={
                "stage": self.stage,
            },
        ).json()

    def create_generator_request(
        self,
        endpoint: str,
        method: HttpMethods,
        content_type: ContentType,
        response_key: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
        path: Optional[str] = None,
        subdomain: Optional[str] = "service",
    ) -> Iterator[Dict[str, Any]]:
        """Prepares a request to be sent over the network.

        Adds auth_token to the headers and creates URL using STAGE.
        Returns an iterator which automatically handles pagination and returns a new page at each iteration.
        To be used with list endpoints only, which returns `pageNumber`, `pageCount` keys in response.

        # noqa: DAR201

        Args:
            endpoint (str): The endpoint to send the request to.
            method (HttpMethods): The HTTP method to use.
            content_type (ContentType): The content type of the request.
            response_key (str): key in the response dict which contains actual list data.
            params (Optional[Dict[str, Any]]): params to send to the request.
            body (Optional[Dict[str, Any]]): body to send to the request.
            path (Optional[str]): path to the file or folder that will be compressed and used as artifact.
            subdomain (Optional[str]): Subdomain for the endpoint. Defaults to `service`.

        Yields:
            Iterator[Dict[str, Any]]: paginated response json, element wise.

        Raises:
            StopIteration: There are no more pages to list
        """
        page_number: int = 1
        page_count: int = 1
        params = params or {}
        while page_number <= page_count:
            params = {**params, "pageNumber": page_number}
            response = self.create_request(
                endpoint,
                method,
                content_type,
                params=params,
                body=body,
                path=path,
                subdomain=subdomain,
            )
            page_count = response["pageCount"]
            yield from response[response_key]
            page_number += 1
        return f"No more {response_key} to list"  # type: ignore[return-value]

    def create_download_request(
        self,
        endpoint: str,
        method: HttpMethods,
        content_type: ContentType,
        download_path: str,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Prepares a request to be sent over the network.

        Adds auth_token to the headers and creates URL using STAGE.
        To be used for file download requests.

        Args:
            endpoint (str): The endpoint to send the request to.
            method (HttpMethods): The HTTP method to use.
            content_type (ContentType): The content type of the request.
            download_path (str): Path where the downloaded file will be stored.
            params (Dict[str, Any], optional): params to send to the request, defaults to None
            body (Dict[str, Any], optional): body to send to the request, only used in multipart requests, defaults to None

        Raises:
            InvalidPathException: The download_path is invalid.
        """
        headers: Dict[str, str] = {"Authorization": self.auth_token}
        url: str = f"{self.base_domain}/{endpoint}"
        response: Any = self.handler.make_request(
            url,
            method,
            content_type=content_type,
            headers=headers,
            params=params or {},
            body=body or {},
            request_kwargs={
                "stream": True,
                "allow_redirects": True,
            },
        )
        try:
            with Path(download_path).open("wb") as fd:
                for chunk in response.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                    fd.write(chunk)
        except IsADirectoryError:
            raise exceptions.InvalidPathException(
                download_path,
                "Path should include the name with which the downloaded file should be stored.",
            ) from None

    def _set_base_domain(self) -> None:
        self.base_domain = get_base_domain(stage=self.stage.value)

    def _set_auth_token(self, auth_token: Optional[str]) -> None:
        if auth_token is not None:
            self.auth_token = auth_token
            return

        logger.info("auth_token not given, searching for 'PEAK_AUTH_TOKEN' or 'API_KEY' in env variables")
        if not os.environ.get("API_KEY") and not os.environ.get("PEAK_AUTH_TOKEN"):
            raise exceptions.MissingEnvironmentVariableException(env_var="PEAK_AUTH_TOKEN or API_KEY")
        if os.environ.get("PEAK_AUTH_TOKEN"):
            self.auth_token = os.environ["PEAK_AUTH_TOKEN"]
            return
        self.auth_token = os.environ["API_KEY"]

    def _set_stage(self, stage: Optional[str]) -> None:
        if stage is not None:
            self.stage = Stage(stage)
            return

        logger.info("stage not given, searching for STAGE in env variables")
        if not os.environ.get("STAGE"):
            logger.info("STAGE environment variable is not set, defaulting to PROD")
            self.stage = Stage.PROD
            return
        self.stage = Stage(os.environ["STAGE"])


__all__: List[str] = ["Session", "_get_default_session"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Supported Log handlers."""
from enum import Enum


class LogHandler(Enum):
    """Enumeration of log handlers to be used in logging.

    Each enum member corresponds to a specific handler defined in the logging module.
    This enum provides a convenient way to specify handlers when configuring loggers.

    Attributes:
        CONSOLE: Represents a console handler, intended for displaying logs in the console.
        FILE: Represents a file handler, intended for writing logs to a file.
    """

    CONSOLE = "CONSOLE"
    FILE = "FILE"



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Apps client module."""

from __future__ import annotations

import json
from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.session import Session


class App(BaseClient):
    """Client class for interacting with Apps."""

    SPECS_BASE_ENDPOINT = "v1/apps/specs"
    DEPLOYMENTS_BASE_ENDPOINT = "v1/apps/deployments"

    @overload
    def list_specs(
        self,
        status: Optional[List[str]] = None,
        featured: Optional[bool] = None,
        name: Optional[str] = None,
        title: Optional[str] = None,
        sort: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_specs(
        self,
        status: Optional[List[str]] = None,
        featured: Optional[bool] = None,
        name: Optional[str] = None,
        title: Optional[str] = None,
        sort: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_specs(
        self,
        status: Optional[List[str]] = None,
        featured: Optional[bool] = None,
        name: Optional[str] = None,
        title: Optional[str] = None,
        sort: Optional[List[str]] = None,
        scope: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Lists all published App specs ordered by creation date. Returns info for the latest release for each spec.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/get_v1_apps_specs_>`__

        Args:
            status (List[str] | None): List of statuses to filter specs.
                Valid values are `available`, `unavailable` and `archived`.
            featured (bool | None): Whether to only return featured specs.
            name (str | None): Only return specs whose names begins with the query string.
            title (str | None): Only return specs whose title begins with the query string.
            sort (List[str] | None): List of fields with desired ordering in the format `[<field>:<order>, ...]`,
                where `order` is one of `['asc', 'desc']` and field is an ordered parameter within the response,
                defaults to `[]`. Valid fields are `createdAt`, `createdBy`, `name`, `title`.
            scope (List[str] | None): List of scopes to only return specs of those scopes. Valid values are `private`, `public` and `shared`.
            page_size (int | None): Number of specs to include per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of specs for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
            StopIteration: There are no more pages to list
        """
        method, endpoint = HttpMethods.GET, f"{self.SPECS_BASE_ENDPOINT}/"
        params: Dict[str, Any] = {
            "status": status,
            "featured": featured,
            "name": name,
            "title": title,
            "sort": sort,
            "scope": scope,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="specs",
                params=params,
                subdomain="press",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
            subdomain="press",
        )

    def create_spec(
        self,
        body: Dict[str, Any],
        featured: Optional[bool] = None,
        scope: Optional[str] = None,
        tenants: Optional[List[str]] = None,
    ) -> Dict[str, str]:
        """Creates a new App spec.

        All App Specs must have unique names within a tenant.
        An App Spec must now be made up of already existing block specs.
        Spec details (Block spec ID and Release version) needs to be added to the `config` key.
        `config` also contains a `autoRunOnDeploy` key which is a boolean value to specify if the resource should be executed on successful deployment of the app. By default, it is False.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/post_v1_apps_specs_>`__

        Args:
            body (Dict[str, Any]):  A payload specifying an App's metadata, release and corresponding config in the expected format.
            featured (bool | None): Whether to feature this spec. By default it is False.
            scope (str | None): Specify weather tenants can discover and deploy this spec.
                `private` restricts the spec to this tenant alone, `public` makes it available on all tenants
                and `shared` allows specifying what set of tenants can access and use this spec.
                By default it is `private`.
            tenants (List[str] | None): Given a shared scope, specify what other tenants can discover and deploy this spec.

        Returns:
            Dict[str, str]: Id of the created app spec.

        SCHEMA:
            .. code-block:: json

                {
                    "version": "number(required)",
                    "kind": "string(required)",
                    "metadata": {
                        "name": "string(required)",
                        "title": "string",
                        "summary": "string(required)",
                        "description": "string",
                        "descriptionContentType": "string",
                        "imageUrl": "string",
                        "tags": [
                            {
                                "name": "string",
                            }
                        ]
                    },
                    "release": {
                        "notes": "string(required)",
                        "version": "string(required)",
                    },
                    "config": [
                        {
                            "id": "string(required)",
                            "release": {
                                "version": "string(required)",
                            },
                            "autoRunOnDeploy": "boolean (By default it is False)",
                        }
                    ]
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.SPECS_BASE_ENDPOINT}/"

        body = {
            "spec": json.dumps(body),
            "featured": json.dumps(featured),
            "scope": scope,
        }

        if tenants:
            body["tenants"] = json.dumps(tenants)

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )

    def describe_spec(
        self,
        spec_id: str,
    ) -> Dict[str, Any]:
        """Describes an existing App spec and also provides details of the latest release.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/get_v1_apps_specs__specId_>`__

        Args:
            spec_id (str): The ID of the app spec to retrieve.

        Returns:
            Dict[str, Any]: Dictionary containing the details of the spec.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given app spec does not exist.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    def update_spec_metadata(self, spec_id: str, body: Dict[str, Any]) -> Dict[None, None]:
        """Updates the metadata of an App spec.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/patch_v1_apps_specs__specId_>`__

        Args:
            spec_id (str): ID of the spec to update
            body (Dict[str, Any]): Dictionary containing the new spec metadata

        Returns:
            dict: Dictionary containing the details of the updated spec.

        SCHEMA:
            .. code-block:: json

                {
                    "metadata": {
                        "name": "string",
                        "title": "string",
                        "summary": "string",
                        "description": "string",
                        "descriptionContentType": "string",
                        "imageUrl": "string",
                        "tags": [
                            {
                                "name": "string",
                            }
                        ],
                        "status": "string",
                    },
                    "featured": "boolean",
                    "scope": "string",
                    "tenants": [
                        "string",
                    ]
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App spec does not exist.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.PATCH, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )

    def create_spec_release(self, spec_id: str, body: Dict[str, Any]) -> Dict[str, str]:
        """Publish a new release to an existing App spec.

        Spec details (Block spec ID and Release version) needs to be added to the `config` key.
        `config` also contains a `autoRunOnDeploy` key which is a boolean value to specify if the resource should be executed on successful deployment of the app. By default, it is False.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/post_v1_apps_specs__specId__releases>`__

        Args:
            spec_id (str): Id of the spec in which new release will be created
            body (Dict[str, Any]): Dictionary containing updated release and config in the expected format.

        Returns:
            Dict[str, str]: Dictionary containing spec id and release version

        SCHEMA:
            .. code-block:: json

                {
                    "config": [
                        {
                            "id": "string(required)",
                            "release": {
                                "version": "string(required)",
                            },
                            "autoRunOnDeploy": "boolean (By default it is False)",
                        }
                    ],
                    "release": {
                        "notes": "string(required)",
                        "version": "string(required)",
                    }
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given app spec does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}/releases"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )

    def delete_spec(
        self,
        spec_id: str,
    ) -> Dict[None, None]:
        """Delete an App Spec and all it's associated releases.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/delete_v1_apps_specs__specId_>`__

        Args:
            spec_id (str): The ID of the App spec to delete.

        Returns:
            dict: Empty dict object.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App spec does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    def describe_spec_release(self, spec_id: str, version: str) -> Dict[str, Any]:
        """Describes an existing App spec release.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/get_v1_apps_specs__specId__releases__release_>`__

        Args:
            spec_id (str): The ID of the app spec to retrieve.
            version (str): The release version of spec to retrieve in valid semantic versioning format.

        Returns:
            Dict[str, Any]: Dictionary containing details of the spec release

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given app spec does not exist.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}/releases/{version}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    @overload
    def list_spec_releases(
        self,
        spec_id: str,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_spec_releases(
        self,
        spec_id: str,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_spec_releases(
        self,
        spec_id: str,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Get all releases of an App spec (ordered by most recently created to oldest).

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Specs/get_v1_apps_specs__specId__releases>`__

        Args:
            spec_id (str): The ID of the app spec to retrieve.
            sort (List[str] | None): List of fields with desired ordering in the format `[<field>:<order>, ...]`,
                where `order` is one of `['asc', 'desc']` and field is an ordered parameter within the response.
                Valid fields are `createdAt` and `createdBy`.
            page_size (int | None, optional): Number of releases to include per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of releases for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App spec does not exist.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
            StopIteration: There are no more pages to list
        """
        method, endpoint = HttpMethods.GET, f"{self.SPECS_BASE_ENDPOINT}/{spec_id}/releases"
        params: Dict[str, Any] = {"pageSize": page_size, "sort": sort}

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="releases",
                params=params,
                subdomain="press",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
            subdomain="press",
        )

    def create_deployment(self, body: Dict[str, str]) -> Dict[str, str]:
        """Creates a new deployment from App spec.

        Uses the latest spec release if release version is not provided.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments>`__

        Args:
            body (Dict[str, str]): Dictionary containing deployment metadata, spec ID, release version, revision info and parameters that would used at run-time.

        Returns:
            Dict[str, str]: ID of the created App deployment

        SCHEMA:
            .. code-block:: json

                {
                    "metadata": {
                        "description": "string"
                        "descriptionContentType": "string"
                        "imageUrl": "string"
                        "name": "string(required)",
                        "summary": "string(required)",
                        "tags": [
                            {
                                "name": "string"
                            }
                        ],
                        "title": "string"
                    },
                    "parameters": {
                        "spec_name": {
                            "build": {
                                "param_name": "param_value (string | number | boolean | array)"
                            },
                            "run": {
                                "param_name": "param_value (string | number | boolean | array)"
                            }
                        },
                    },
                    "revision": {
                        "notes": "string(required)",
                    },
                    "spec": {
                        "id": "string(required)",
                        "release": {
                            "version": "string(required)",
                        }
                    }
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given app spec release does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )

    @overload
    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_deployments(
        self,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Lists App deployments ordered by creation date.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments>`__

        Args:
            status (List[str] | None): List of statuses to filter deployments.
                Valid values are `deploying`, `deployed`, `deleting`, `delete_failed`, `failed`, `platform_resource_error`, `redeploying`, `rollback`, `rollback_complete`, `rollback_failed`, and `warning`.
            name (str | None): Only return deployments whose names begins with the query string.
            sort (List[str] | None): List of fields with desired ordering in the format `[<field>:<order>, ...]`,
                where `order` is one of `['asc', 'desc']` and field is an ordered parameter within the response.
                Valid fields are `createdAt`, `createdBy`, `name` and `title`.
            page_size (int | None): Number of deployments to include per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of deployments for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
            StopIteration: There are no more pages to list.
        """
        method, endpoint = HttpMethods.GET, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/"

        params: Dict[str, Any] = {
            "status": status,
            "name": name,
            "sort": sort,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="deployments",
                params=params,
                subdomain="press",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
            subdomain="press",
        )

    def describe_deployment(self, deployment_id: str) -> Dict[str, Any]:
        """Describes an existing App deployment, also provides details of the latest revision.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments__deploymentId_>`__

        Args:
            deployment_id (str): The ID of the App deployment to retrieve.

        Returns:
            Dict[str, Any]: Dictionary containing the details of the deployment.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment does not exist.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    def redeploy(self, deployment_id: str) -> Dict[str, Any]:
        """Redeploy latest revision of an existing App deployment.

        This function allows you to redeploy an App deployment that is in a `failed` or `warning` state, provided at least one of its block deployments is also in a `failed` or `warning` state.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments__deploymentId__redeploy>`__

        Args:
            deployment_id (str): The ID of the App deployment to redeploy.

        Returns:
            Dict[str, Any]: A dictionary containing details of the deployment.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment or its revisions do not exist.
            ConflictException: There is a conflict with the current state of the target resource.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}/redeploy"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    def delete_deployment(self, deployment_id: str) -> Dict[None, None]:
        """Deletes an App deployment.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/delete_v1_apps_deployments__deploymentId_>`__

        Args:
            deployment_id (str): The ID of the App deployment to delete.

        Returns:
            dict: Empty dict object.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    def create_deployment_revision(self, deployment_id: str, body: Dict[str, str]) -> Dict[str, str]:
        """Creates a new Revision of a Deployment, from a selected App Spec Release.

        Publish a new revision of an App Deployment. The payload must specify the release version of the parent spec that you wish to use (if one is not provided, the latest available will be used),
        optional revision notes in the expected format,, and optional parameters if required by the associated block spec releases.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/post_v1_apps_deployments__deploymentId__revisions>`__

        Args:
            deployment_id (str): Id of the deployment in which new revision will be created.
            body (Dict[str, str]): release version, revision notes, and any parameters required

        Returns:
            Dict[str, str]: ID of the created App deployment, and the new Revision number

        SCHEMA:
            .. code-block:: json

                {
                    "release": {
                        "version": "string"
                    },
                    "revision": {
                        "notes": "string",
                    },
                    "parameters": {
                        "block-name-1":{
                            "build": {
                                "param_name": "param_value (string | number | boolean | array)"
                            },
                            "run": {
                                "param_name": "param_value (string | number | boolean | array)"
                            }
                        },
                        "block-name-2":{
                            "build": {
                                "param_name": "param_value (string | number | boolean | array)"
                            },
                            "run": {
                                "param_name": "param_value (string | number | boolean | array)"
                            }
                        }
                    }
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given deployment or spec release does not exist.
            ConflictException: There is a conflict with the current state of the target resource.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}/revisions"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )

    def describe_deployment_revision(self, deployment_id: str, revision: str) -> dict[str, Any]:
        """Describes an existing App deployment revision, Parameters listed in the response are masked if "hideValue" was set to true when creating the associated block specs.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/get_v1_apps_deployments__deploymentId__revisions__revision_>`__

        Args:
            deployment_id (str): The ID of the App deployment to retrieve the revision from.
            revision (str): The revision number to retrieve

        Returns:
            Dict[str, Any]: Dictionary containing the details of the deployment revision.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment or revision does not exist.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}/revisions/{revision}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            subdomain="press",
        )

    @overload
    def list_deployment_revisions(
        self,
        deployment_id: str,
        sort: Optional[List[str]] = None,
        status: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_deployment_revisions(
        self,
        deployment_id: str,
        sort: Optional[List[str]] = None,
        status: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_deployment_revisions(
        self,
        deployment_id: str,
        sort: Optional[List[str]] = None,
        status: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Lists revisions for an App deployment ordered by creation date or provided sort key.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments__deploymentId__revisions>`__

        Args:
            deployment_id (str): The ID of the App deployment to retrieve the revision from.
            sort (List[str] | None): List of fields with desired ordering in the format `[<field>:<order>, ...]`,
                where `order` is one of `['asc', 'desc']` and field is an ordered parameter within the response.
                Valid fields are `createdAt`, `createdBy`, `name` and `title`.
            status (List[str] | None): List of statuses to filter revisions.
                Valid values are `deleting`, `delete_failed`, `deployed`, `deploying`, `failed`, `platform_resource_error`, `rollback`, `rollback_complete`, `rollback_failed` and `superseded`.
            page_size (int | None): Number of revisions to include per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of revisions for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server encountered an unexpected condition that
                prevented it from fulfilling the request.
            StopIteration: There are no more pages to list.
        """
        method, endpoint = HttpMethods.GET, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}/revisions"

        params: Dict[str, Any] = {
            "sort": sort,
            "status": status,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="deployments",
                params=params,
                subdomain="press",
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
            subdomain="press",
        )

    def update_deployment_metadata(self, deployment_id: str, body: Dict[str, Dict[str, str]]) -> Dict[None, None]:
        """Update the metadata of an App deployment.

        REFERENCE:
            🔗 `API Documentation <https://press.peak.ai/api-docs/index.htm#/App%20Deployments/patch_v1_apps_deployments__deploymentId_>`__

        Args:
            deployment_id (str): ID of the App deployment to update.
            body (Dict[str, Dict[str, str]]): Dictionary of the new deployment metadata.

        Returns:
            dict: Details of the updated deployment.

         SCHEMA:
            .. code-block:: json

                {
                    "description": "string",
                    "descriptionContentType": "string",
                    "imageUrl": "string",
                    "name": "string",
                    "summary": "string",
                    "tags": [
                        {
                        "name": "string",
                        }
                    ],
                    "title": "string",
                }

        Raises:
            BadRequestException: The given parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given App deployment does not exist.
            InternalServerErrorException: the server encountered an unexpected condition that
                prevented it from fulfilling the request.
        """
        method, endpoint = HttpMethods.PATCH, f"{self.DEPLOYMENTS_BASE_ENDPOINT}/{deployment_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
            subdomain="press",
        )


def get_client(session: Optional[Session] = None) -> App:
    """Returns an App client.

    Args:
        session (Optional[Session]): A Session Object. If no session is provided, a default session is used.

    Returns:
        App: the App client object
    """
    return App(session)


__all__: List[str] = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Compression module to create zip file to be used as artifact."""

from __future__ import annotations

import contextlib
import os
import tempfile
import zipfile
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, Optional, Set

from pathspec import PathSpec

from peak import constants, exceptions, output

ZIP_COMPRESSION = zipfile.ZIP_DEFLATED


@contextlib.contextmanager
def compress(path: str, ignore_files: Optional[list[str]] = None) -> Iterator[tempfile.SpooledTemporaryFile[bytes]]:
    """Creates compressed zip of the files in path.

    Args:
        path (str): Path of the folder to create zip file.
        ignore_files (Optional[list[str]]): Ignorefiles to use when creating zip file.

    Yields:
        Iterator[tempfile.SpooledTemporaryFile[bytes]]: Bytes of the zip file in chunks.

    Raises:
        InvalidPathException: Given path is invalid and cannot be traversed.
        FileLimitExceededException: Compressed directory size exceeded the max limits.
    """
    path_obj = Path(path)
    if not path_obj.is_dir():
        raise exceptions.InvalidPathException(path, "Either the path does not exist or is not a directory.")
    included_files = get_files_to_include(path, ignore_files)
    parent_directories: Set[Path] = set()

    with tempfile.SpooledTemporaryFile() as tmp_file:
        with zipfile.ZipFile(tmp_file, "w", compression=ZIP_COMPRESSION) as zf:
            for file in included_files:
                zf.write(path_obj / file, file)
                parent_directories.update(Path(file).parents)
                if tmp_file.tell() > constants.MAX_ARTIFACT_SIZE_MB * constants.MB:
                    raise exceptions.FileLimitExceededException(constants.MAX_ARTIFACT_SIZE_MB)

            # include directories as API backend need directories explicitly included
            relative_root_path = Path()
            if relative_root_path in parent_directories:
                parent_directories.remove(relative_root_path)
            for directory in parent_directories:
                zf.write(path_obj / directory, directory)

        tmp_file.seek(0)
        yield tmp_file


def print_zip_content(zip_file: tempfile.SpooledTemporaryFile[bytes]) -> None:
    """Prints content of the zip file to stdout.

    Args:
        zip_file (BinaryIO): Opened zip file in binary format.
    """
    with zipfile.ZipFile(zip_file, "r") as zf:
        zf.printdir()


def get_files_to_include(path: str, ignore_files: Optional[list[str]] = None) -> Iterator[str]:
    """Get a list of file paths to be included.

    Args:
        path (str): Root path from where to start the traversal.
        ignore_files (Optional[list[str]]): Path of ignorefiles to use, defaults to .dockerignore.

    Yields:
        Iterator[str]: Path of files to be included.
    """
    path_obj = Path(path)
    ignore_pathspec = _load_ignore_patterns(path_obj, ignore_files)
    yield from ignore_pathspec.match_tree(path, follow_links=True)


def _reverse_pattern(pattern: str) -> str:
    """Converts the inclusion pattern to exclusion and vice-versa.

    This is needed because PathSpec will return the files that matches the patterns
    but we want the opposite, i.e. we want to exclude the files which match the
    patterns in ignorefiles.

    Args:
        pattern (str): Pattern to process

    Returns:
        str: Processed pattern
    """
    pattern = pattern.strip()
    if not pattern or pattern.startswith("#"):
        return pattern
    if pattern.startswith("!"):
        return pattern[1:]
    return "!" + pattern


def _load_ignore_patterns(path_obj: Path, ignore_files: Optional[list[str]]) -> PathSpec:
    """Reads ignorefiles and loads all patterns into PathSpec.

    Args:
        path_obj (Path): Root path relative to which ignore_files would be searched.
        ignore_files (list[str]): Path of ignorefiles relative to where the script is running

    Returns:
        PathSpec: PathSpec object with all patterns

    Raises:
        InvalidPathException: Given path is invalid and cannot be traversed.
    """
    if ignore_files is None or len(ignore_files) == 0:
        # use .dockerignore as default if no ignore file is provided
        default_ignore = path_obj / ".dockerignore"
        ignore_files = [os.fspath(default_ignore)] if default_ignore.exists() else []

    all_patterns: list[str] = [
        "*",
    ]  # include everything, apply ignore patterns on top of it, if this is given nothing will be included
    for ignore_file in ignore_files:
        ignore_file_path = path_obj / ignore_file
        normalized_ignore_path = ignore_file_path.resolve().relative_to(path_obj.resolve())
        # ignorefiles should only be at root level
        if len(normalized_ignore_path.parents) != 1:
            raise exceptions.InvalidPathException(
                ignore_file,
                "Ignore file should be present at root level of given path.",
            )

        patterns = ignore_file_path.open("r").read().splitlines()
        all_patterns.extend(
            map(
                _reverse_pattern,
                patterns,
            ),
        )
    return PathSpec.from_lines("gitwildmatch", all_patterns)


def print_file_tree(files: Iterable[str]) -> None:
    """Prints list of files in tree format with specific limits per level.

    Args:
        files (list[str]): List of file paths
    """
    writer = output.Writer(ignore_debug_mode=True)
    files_dict = _build_files_dict(files)

    limits = {1: 100, 2: 100, 3: 50}
    default_limit = 25

    def _print_tree(files_dict: Dict[str, Any], indent: str, level: int) -> None:
        limit = limits.get(level, default_limit)
        count = 0

        for key, value in files_dict.items():
            if isinstance(value, dict):
                writer.write(f"{indent}{key}/")
                new_indent = indent + ("|   " if indent else "├── ")
                _print_tree(value, new_indent, level + 1)

        for key, value in files_dict.items():
            if not isinstance(value, dict):
                if count >= limit:
                    writer.write(f"{indent}...")
                    break
                writer.write(f"{indent}{key}")
                count += 1

    _print_tree(files_dict, "", 1)


def _build_files_dict(files: Iterable[str]) -> Dict[str, Any]:
    """Builds a nested dictionary from list of files.

    Args:
        files (list[str]): List of file paths to process.

    Returns:
        dict[str, Any]: Nested dict file tree structure.
    """
    files_dict: Dict[str, Any] = {}
    for f in files:
        components = Path(os.path.normpath(f)).parts
        current_dir = files_dict
        for directory in components[:-1]:
            if directory not in current_dir or current_dir[directory] is None:
                current_dir[directory] = {}
            current_dir = current_dir[directory]
        if components[-1] not in current_dir:
            current_dir[components[-1]] = None
    return files_dict



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Peak Metric commands."""

import json
from typing import Any, Dict, List, Optional

import typer
from peak import config
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.helpers import combine_dictionaries, parse_list_of_strings, variables_to_dict
from peak.metrics.metrics import Metric
from peak.output import Writer
from typing_extensions import Annotated

app = typer.Typer(
    help="Metrics commands.",
    short_help="Manage Metrics.",
)

_ARTIFACT_PATH = typer.Option(None, help="Path to the artifact.")

_PUBLISH_NAMESPACE = _NAMESPACE = typer.Option(
    None,
    help="The namespace where you intend to publish the metrics. If not provided, the default namespace is used.",
)

_NAMESPACE = typer.Option(
    None,
    help="The namespace associated with the metrics. If not provided, the default namespace is used.",
)
_GENERATE_SQL = typer.Option(
    None,
    help="Indicates whether to return the SQL query instead of data. If `true`, the response will include the SQL query used to retrieve the metrics. Default is `false`.",
)
_MEASURES = typer.Option(
    None,
    help="An array of measures to include in the query. Measures represent quantitative metrics such as sums or counts. Provide them in stringified JSON format.",
)
_DIMENSIONS = typer.Option(
    None,
    help="An array of dimensions to include in the query. Dimensions represent qualitative categories such as time, location, or product names. Provide them in stringified JSON format.",
)
_FILTERS = typer.Option(
    None,
    help="An array of filter objects to apply to the query. Filters limit the data returned based on specific conditions. Provide them in stringified JSON format.",
)
_TIME_DIMENSIONS = typer.Option(
    None,
    help="An array of time dimensions to include in the query. Time dimensions allow querying over specific time ranges with optional granularity (e.g., day, month, year). Provide them in stringified JSON format.",
)
_SEGMENTS = typer.Option(
    None,
    help="An array of segments to include in the query. Segments represent pre-defined filters that can be applied to metrics. Provide them in stringified JSON format.",
)
_ORDER = typer.Option(
    None,
    help="Defines the sort order of the results. This is an object where keys are the dimensions/measures and values are either 'asc' or 'desc' to specify ascending or descending order. Provide them in stringified JSON format.",
)
_LIMIT = typer.Option(
    None,
    help="Limits the number of rows returned by the query. If not provided, the default limit is applied.",
)
_OFFSET = typer.Option(
    None,
    help="Specifies the number of rows to skip before starting to return data. Useful for pagination.",
)

_METRIC_TYPES = typer.Option(
    None,
    help="The type of metric to create. If not provided, all metrics are retrieved. Available types are `cube`, `view`, `dimension`, `measure`, `segment` and `all`.",
)

_COLLECTION_NAME = typer.Option(
    None,
    help="Name of the metric collection.",
)
_PUBLICATION_ID = typer.Option(
    None,
    help="The publication ID associated with the metrics.",
)

_MEASURES_DELETE = typer.Option(
    None,
    help="An array of measures to delete.",
)

_COLLECTION_ID = typer.Option(
    None,
    help="The ID of the collection to publish the metrics.",
)

_COLLECTION_SCOPE = typer.Option(
    None,
    help="Scope of the metrics collection. Must be one of the following: PUBLIC, PRIVATE.",
)

_COLLECTION_DESCRIPTION = typer.Option(
    None,
    help="Description of the metric collection.",
)

_COLLECTION_IDS = typer.Option(
    None,
    help="An array of collection IDs to include in the query. If not provided, all collections are retrieved.",
)

_COLLECTION_SCOPES = typer.Option(
    None,
    help="An array of scopes to filter the collections by. Available scopes are `PUBLIC` and `PRIVATE`. If not provided, all collections of the tenant along with all public collections are retrieved.",
)


@app.command(short_help="Publish metrics.")
def publish(
    ctx: typer.Context,
    file: Annotated[
        Optional[str],
        typer.Argument(
            ...,
            help="Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template.",
        ),
    ] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    namespace: Optional[str] = _PUBLISH_NAMESPACE,
    artifact_path: Optional[str] = _ARTIFACT_PATH,
    collection_id: Optional[str] = _COLLECTION_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Publish*** metrics.

    The metrics can be published either by passing artifact and namespace,
    or by passing collection_id and namespace. If both artifact and collection_id
    are provided, artifact takes priority. If the namespace is not provided, the 'default' namespace is used.

    \b
    🧩 ***Input file schema(yaml):***<br/>

     **_Publish metrics using artifact and namespace_**
    ```yaml
      body (map):
        namespace (str): The namespace associated with the metrics.
      artifact (map):
        path (str): Path to the artifact.
    ```
     **_Publish metrics using collection id and namespace_**
    ```yaml
      body (map):
        namespace (str): The namespace associated with the metrics.
      collectionId (str): The ID of the collection to publish the metrics.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak metrics publish '/path/to/metrics.yaml' --params-file '/path/to/values.yaml'
    ```

    We can also provide the required parameters from the command line or combine the YAML template and command line arguments in which case the command line arguments will take precedence.

    \b
    📝 ***Example usage without yaml:***
    ```bash
    # Publish metrics using artifact and namespace
    peak metrics publish --artifact-path <path> --namespace <namespace>
    # Publish metrics using collection id and namespace
    peak metrics publish --collection-id <collection-id> --namespace <namespace>
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "artifactId": "7dc0feaa-be90-467b-9c3a-009a234e4b2b",
        "collectionId": "bc8b6ef5-d2f6-4b7f-9365-0261b43997c9",
        "publicationId": "79af8462-2820-483c-a8b6-d697555a8fc2",
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/post_api_v1_metrics)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        namespace,
    )

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, metrics_client.publish)

    updated_body = combine_dictionaries(body.get("body") or {}, user_options)
    artifact = helpers.get_updated_artifacts(body, artifact_path, None)

    with writer.pager():
        response: Dict[str, Any] = metrics_client.publish(
            body=updated_body,
            artifact=artifact,  # type: ignore  # noqa: PGH003
            collection_id=collection_id,
        )
        writer.write(response)


@app.command(short_help="Query metrics.")
def query(
    ctx: typer.Context,
    file: Annotated[
        Optional[str],
        typer.Argument(
            ...,
            help="Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template.",
        ),
    ] = None,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    namespace: Optional[str] = _NAMESPACE,
    generate_sql: Optional[bool] = _GENERATE_SQL,
    measures: Optional[List[str]] = _MEASURES,
    dimensions: Optional[List[str]] = _DIMENSIONS,
    filters: Optional[List[str]] = _FILTERS,
    time_dimensions: Optional[List[str]] = _TIME_DIMENSIONS,
    segments: Optional[List[str]] = _SEGMENTS,
    order: Optional[str] = _ORDER,
    limit: Optional[int] = _LIMIT,
    offset: Optional[int] = _OFFSET,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Query*** a published metric in the semantic layer using the provided parameters.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
        namespace (str | required: false): The namespace associated with the metrics. If not provided, the default namespace is used.
        generateSql (bool | required: false): Indicates whether to return the SQL query instead of data. If `true`, the response will include the SQL query used to retrieve the metrics. Default is `false`.
        measures (list(str) | required: false): An array of measures to include in the query. Measures represent quantitative metrics such as sums or counts.
        dimensions (list(str) | required: false): An array of dimensions to include in the query. Dimensions represent qualitative categories such as time, location, or product names.
        filters (list(map) | required: false):
            dimension (str): The dimension to filter on. Supported values are `equals`, `notEquals`, `contains`, `notContains`, `startsWith`, `notStartsWith`, `endsWith`, `notEndsWith`, `gt`, `gte`, `lt`, `lte`, `inDateRange`, `notInDateRange`, `beforeDate`, `beforeOrOnDate`, `afterDate`, `afterOrOnDate` etc.
            operator (str): The operator to use for the filter.
            values (list(str)): An array of values to filter on.
        timeDimensions (list(map) | required: false):
            dimension (str): The time dimension to include in the query.
            granularity (str | required: false): The granularity of the time dimension. Supported values are `second`, `minute`, `hour`, `day`, `week`, `month`, `quarter`, and `year`.
            dateRange (list(str) | str | required: false): An array of two dates that define the time range for the query. Alternatively, you can provide a single string out of the following predefined date ranges `today`, `yesterday`, `this week`, `last week`, `this month`, `last month`, `this quarter`, `last quarter`, `this year`, `last year`, `last 7 days` and `last 30 days`.
        segments (list(str) | required: false): An array of segments to include in the query. Segments represent pre-defined filters that can be applied to metrics.
        order (map | required: false): Defines the sort order of the results. This is a stringified object where keys are the dimensions/measures and values are either 'asc' or 'desc' to specify ascending or descending order.
        limit (int | required: false): Limits the number of rows returned by the query. If not provided, the default limit is applied.
        offset (int | required: false): Specifies the number of rows to skip before starting to return data. Useful for pagination.
    ```

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak metrics query '/path/to/query.yaml' --params-file '/path/to/values.yaml'
    ```

    We can also provide the required parameters from the command line or combine the YAML template and command line arguments in which case the command line arguments will take precedence.

    \b
    📝 ***Example usage without yaml:***<br/>
    ```bash
    peak metrics query --measures "<cube_name>.<resource_name_1>" --measures "<cube_name>.<resource_name_2>" --dimensions "<cube_name>.<resource_name>" --time-dimensions "{\"dimension\":\"<cube_name>.<resource_name>\",\"dateRange\":[\"2024-01-26T00:00:00Z\",\"2024-06-06T00:00:00Z\"],\"granularity\":\"day\"}"
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "data": [
            {
                "region": "North America",
                "total_sales": 1000000
            },
            {
                "region": "Europe",
                "total_sales": 500000
            }
        ],
        "sql": "SELECT region, SUM(total_sales) AS total_sales FROM orders GROUP BY region"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Query/get_api_v1_metrics_query)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    request_params: Dict[str, Any] = {}

    if file:
        request_params = helpers.template_handler(file, params_file, params)
        request_params = helpers.remove_unknown_args(request_params, metrics_client.query)

    cli_params = {
        "namespace": namespace,
        "generate_sql": generate_sql,
        "measures": parse_list_of_strings(measures) if measures else [],
        "dimensions": parse_list_of_strings(dimensions) if dimensions else [],
        "filters": [json.loads(query_filter) for query_filter in filters] if filters else [],
        "time_dimensions": (
            [json.loads(time_dimension) for time_dimension in time_dimensions] if time_dimensions else []
        ),
        "segments": parse_list_of_strings(segments) if segments else [],
        "order": json.loads(order) if order else {},
        "limit": limit,
        "offset": offset,
    }

    merged_params = {**request_params}
    for key, value in cli_params.items():
        if (value not in [None, [], {}] and key in merged_params) or key not in merged_params:
            merged_params[key] = value

    with writer.pager():
        response = metrics_client.query(**merged_params)
        cli_output_type = config.OUTPUT_TYPE

        if "data" in response:
            total_count = len(response["data"])
            response["totalCount"] = total_count

        if "generate_sql" in merged_params and merged_params["generate_sql"] is True and cli_output_type == "table":
            response["data"] = []

        writer.write(response)


@app.command("list", short_help="List metrics.")
def list_metrics(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    namespace: Optional[str] = _NAMESPACE,
    type: Optional[str] = _METRIC_TYPES,  # noqa: A002
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** metrics in the semantic layer.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak metrics list --page-size 25 --page-number 1 --namespace <namespace> --type <type>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "data": [
            {
                "name": "product",
                "type": "cube",
                "public": true,
                "measures": [
                    {
                        "name": "product.max_price",
                        "type": "number",
                        "aggType": "max",
                        "public": true
                    },
                    {
                        "name": "product.max_discount",
                        "type": "number",
                        "aggType": "max",
                        "public": true
                    },
                ],
                "dimensions": [
                    {
                        "name": "product.sale",
                        "type": "string",
                        "public": true,
                        "primaryKey": true
                    },
                    {
                        "name": "order.created_at",
                        "type": "time",
                        "public": true,
                        "primaryKey": false
                    },
                ],
                "segments": [],
                "collectionId": "b8d8308e-4f45-42e5-9c08-d4c4ba6db7f6",
                "publicationId": "299e7d07-db2f-4050-88c3-40afd7603807"
            }
        ],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25,
        "totalCount": 2
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/get_api_v1_metrics)
    """
    metric_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = metric_client.list(
            page_size=page_size,
            page_number=page_number,
            namespace=namespace,
            type=type,
            return_iterator=False,
        )

        writer.write(response)


@app.command(short_help="Create metrics collection.")
def create_collection(
    ctx: typer.Context,
    file: Annotated[
        Optional[str],
        typer.Argument(
            ...,
            help="Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template.",
        ),
    ] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    name: str = _COLLECTION_NAME,
    scope: str = _COLLECTION_SCOPE,
    description: Optional[str] = _COLLECTION_DESCRIPTION,
    artifact_path: str = _ARTIFACT_PATH,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** metrics collection.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (str): Name of the metric collection.
        scope (str): Scope of the metrics artifact.
        description (str | required: false): Description of the metric collection.
      artifact (map):
        path (str): Path to the artifact.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak metrics create-collection '/path/to/metrics.yaml' --params-file '/path/to/values.yaml'
    ```

    We can also provide the required parameters from the command line or combine the YAML template and command line arguments in which case the command line arguments will take precedence.

    \b
    📝 ***Example usage without yaml:***
    ```bash
    peak metrics create-collection --artifact-path <path> --name <name> --scope <scope> --description <description>
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "artifactId": "7dc0feaa-be90-467b-9c3a-009a234e4b2b",
        "collectionId": "bc8b6ef5-d2f6-4b7f-9365-0261b43997c9",
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/post_api_v1_metrics_collections)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        name,
        scope,
        description,
    )

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, metrics_client.create_collection)

    updated_body = combine_dictionaries(body.get("body") or {}, user_options)
    artifact = helpers.get_updated_artifacts(body, artifact_path, None)

    with writer.pager():
        response: Dict[str, Any] = metrics_client.create_collection(body=updated_body, artifact=artifact)  # type: ignore  # noqa: PGH003
        writer.write(response)


@app.command(short_help="Delete the metrics.")
def delete(
    ctx: typer.Context,
    namespace: Optional[str] = _NAMESPACE,
    measures: Optional[List[str]] = _MEASURES_DELETE,
    publication_id: Optional[str] = _PUBLICATION_ID,
    *,
    dry_run: bool = DRY_RUN,  # noqa: ARG001
    output_type: OutputTypesNoTable = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** one or more measures.

    Measures can either be deleted by passing a namespace and a list of measures or
    by giving a publication id which would delete all the measures associated with that publication. If both are passed,
    publication id takes priority.

    \b
    📝 ***Example usage:***
    ```bash
    # Delete using namespace and measure names
    peak metrics delete --namespace <namespace> --measures <measure> --measures <measure>
    # Delete using publication id
    peak metrics delete --publication-id
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/delete_api_v1_metrics_publications__publicationId_)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = {
        "namespace": namespace,
        "measures": parse_list_of_strings(measures) if measures else [],
        "publication_id": publication_id,
    }

    with writer.pager():
        response: Dict[str, Any] = metrics_client.delete(**body)
        writer.write(response)


@app.command(short_help="Delete metrics collection.")
def delete_collection(
    ctx: typer.Context,
    collection_id: str = typer.Argument(..., help="ID of the metric collection."),
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** metrics collection.

    \b
    📝 ***Example usage:***
    ```bash
    peak metrics delete-collection <collection_id>
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/delete_api_v1_metrics_collections__collectionId_)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response: Dict[str, Any] = metrics_client.delete_collection(collection_id=collection_id)
        writer.write(response)


@app.command(short_help="List metrics collections.")
def list_collections(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    id: Optional[List[str]] = _COLLECTION_IDS,  # noqa: A002
    scope: Optional[List[str]] = _COLLECTION_SCOPES,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** metrics collections.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak metrics list-collections --page-size 25 --page-number 1 --id <collection_id_1> --id <collection_id_2> --scope <scope>
    ```

    \b
    🆗 ***Response:***<br/>
    ```json
    {
        "collections": [
            {
                "id": "b8d8308e-4f45-42e5-9c08-d4c4ba6db7f6",
                "name": "product",
                "tenant": "tenant-name",
                "scope": "PUBLIC",
                "description": "Product metrics",
                "artifactId": "7dc0feaa-be90-467b-9c3a-009a234e4b2b",
                "createdAt": "2024-01-26T00:00:00Z",
                "createdBy": "someone@peak.ai",
                "updatedAt": "2024-01-26T00:00:00Z",
                "updatedBy": "someone@peak.ai"
            }
        ],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25,
        "totalCount": 1
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/get_api_v1_metrics_collections)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = metrics_client.list_collections(
            page_size=page_size,
            page_number=page_number,
            id=parse_list_of_strings(id) if id else [],
            scope=parse_list_of_strings(scope) if scope else [],
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="List namespaces.")
def list_namespaces(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** namespaces.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak metrics list-namespaces --page-size 25 --page-number 1
    ```

    \b
    🆗 ***Response:***<br/>
    ```json
    {
        "namespaces": [
            {
                "name": "default",
                "models": [
                    {
                        "name": "stocks",
                        "publicationId": "5503a4df-fa33-4932-9f60-9e3930f37f65",
                        "type": "cube"
                    }
                ]
            }
        ],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25,
        "totalCount": 1
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/semantic-layer/api-docs/index.htm#/Namespaces/get_api_v1_namespaces)
    """
    metrics_client: Metric = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = metrics_client.list_namespaces(
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

#

"""Peak Webapps commands.

.. warning::
    The ``Web Apps`` commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the ``Service`` commands, which offers all the features of ``Webapp`` along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the ``Service`` commands will ensure easier adoption of its advanced features and future updates.

    For guidance on migrating to the ``Service`` and taking full advantage of its capabilities, please refer to the ``Service`` documentation.
"""

from typing import Any, Dict, List, Optional

import typer
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.helpers import combine_dictionaries, map_user_options, parse_list_of_strings, variables_to_dict
from peak.output import Writer
from peak.resources.webapps import Webapp
from typing_extensions import Annotated

app = typer.Typer(
    help="""Visualize your data and predictions using Webapps.
    ```yaml
    NOTE: Please note that only generic (EKS-based) webapps are supported with CLI. **Note** that, the functionalities provided through the `Webapp` feature are deprecated and will eventually be removed. We recommend using the `Service` feature which offers a more comprehensive and flexible solution for managing web applications and API deployments.
    ```
    """,
    short_help="Create and Manage Webapps.",
)

_WEBAPP_ID = typer.Argument(..., help="ID of the webapp to be used in this operation")

_LIST_STATUS = typer.Option(
    None,
    help="A list of webapp status to filter the list by. Valid values are `CREATING`, `DEPLOYING`, `AVAILABLE`, `DELETING`, `CREATE_FAILED`, `DELETE_FAILED`.",
)

_LIST_NAME = typer.Option(None, help="Webapp name to search for")

FILE_HELP_STRING = (
    "Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template."
)

NAME = typer.Option(None, help="Name of the webapp")

DESCRIPTION = typer.Option(None, help="Description of the webapp")

TITLE = typer.Option(None, help="Title of the webapp")

IMAGE_ID = typer.Option(None, help="ID of the image to be used in this operation.")

VERSION_ID = typer.Option(
    None,
    help="ID of the version to be used in this operation. If version-id is not passed, webapp will be created using latest ready version of the image.",
)

INSTANCE_TYPE_ID = typer.Option(None, help="The ID of the instance type to be used in this operation.")

SESSION_STICKINESS = typer.Option(None, help="Enable session stickiness for the webapp.")

MAPPING = {"imageId": "imageDetails", "versionId": "imageDetails", "instanceTypeId": "resources"}

DEPRECATION_MESSAGE = "The Web App commands are deprecated and will eventually be removed. We recommend using the Service feature which offers a more comprehensive and flexible solution for managing web applications and API deployments.\n"


@app.command("list", short_help="List webapps.")
def list_webapps(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    status: Optional[List[str]] = _LIST_STATUS,
    name: Optional[str] = _LIST_NAME,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all webapps that exist for the tenant.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak webapps list --page-size 10 --page-number 1 --status CREATING --name test
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "webappsCount": 1,
        "webapps": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/list-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = webapp_client.list_webapps(
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
            status=parse_list_of_strings(status),
            name=name,
        )
        writer.write(response, deprecation_message=DEPRECATION_MESSAGE)


@app.command(short_help="Create a new webapp.")
def create(
    ctx: typer.Context,
    file: Annotated[Optional[str], typer.Argument(..., help=FILE_HELP_STRING)] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    name: str = NAME,
    title: str = TITLE,
    description: str = DESCRIPTION,
    image_id: int = IMAGE_ID,
    version_id: int = VERSION_ID,
    instance_type_id: int = INSTANCE_TYPE_ID,
    session_stickiness: bool = SESSION_STICKINESS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new webapp and start its deployment.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (string): Name of the webapp.
        title (string | required: false): Title of the webapp.
        description (string | required: false): Description of the webapp.
        imageDetails (map):
            imageId (number): ID of the image.
            versionId (number | required: false): ID of the image version. If versionId is not passed, webapp will be created using latest ready version of the image.
        resources (map | required: false):
            instanceTypeId (number): ID of the instance type.
        sessionStickiness (boolean | required: false): Enable session stickiness for the webapp. Default value is false. Enabling session stickiness will tie each user to a specific server for all their requests.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak webapps create '/path/to/file.yml' -v '/path/to/params.yml'
    ```

    \b
    📝 ***Example usage without yaml:***
    ```bash
    peak webapps create --name <name> --title <title> --description <description> --image-id <image-id> --version-id <version-id> --instance-type-id <instance-type-id> --session-stickiness
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/create-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        name,
        description,
        title,
        image_id,
        version_id,
        instance_type_id,
        session_stickiness,
    )
    user_options = map_user_options(user_options, MAPPING)

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, webapp_client.create_webapp)

    body = combine_dictionaries(body.get("body") or {}, user_options)

    with writer.pager():
        response = webapp_client.create_webapp(body=body)
        writer.write(response, deprecation_message=DEPRECATION_MESSAGE)


@app.command(short_help="Update an existing webapp.")
def update(
    ctx: typer.Context,
    webapp_id: str = _WEBAPP_ID,
    file: Annotated[Optional[str], typer.Argument(..., help=FILE_HELP_STRING)] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    title: str = TITLE,
    description: str = DESCRIPTION,
    image_id: int = IMAGE_ID,
    version_id: int = VERSION_ID,
    instance_type_id: int = INSTANCE_TYPE_ID,
    session_stickiness: bool = SESSION_STICKINESS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** an existing webapp.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    When updating the webapp, it will trigger a redeployment only under specific conditions.
    Redeployment is triggered if you make changes to any of the following parameters: imageId, versionId, instanceTypeId or sessionStickiness.
    However, only modifying the title or description will not trigger a redeployment.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        title (string | required: false): Title of the webapp.
        description (string | required: false): Description of the webapp.
        imageDetails (map | required: false):
            imageId (number): ID of the image.
            versionId (number | required: false): ID of the image version. If versionId is not passed, webapp will be created using latest ready version of the image.
        resources (map | required: false):
            instanceTypeId (number): ID of the instance type.
        sessionStickiness (boolean | required: false): Enable session stickiness for the webapp. Default value is false. Enabling session stickiness will tie each user to a specific server for all their requests.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak webapps update <webapp-id> '/path/to/file.yml' -v '/path/to/params.yml'
    ```

     \b
    📝 ***Example usage without yaml:***
    ```bash
    peak webapps update <webapp-id> --title <title> --description <description> --image-id <image-id> --version-id <version-id> --instance-type-id <instance-type-id> --session-stickiness
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "ab11c21d-1add-45dd-a72e-8c6b83b68dee"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/update-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        description,
        title,
        image_id,
        version_id,
        instance_type_id,
        session_stickiness,
    )
    user_options = map_user_options(user_options, MAPPING)

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, webapp_client.update_webapp)
    body = combine_dictionaries(body.get("body") or {}, user_options)

    with writer.pager():
        response = webapp_client.update_webapp(webapp_id=webapp_id, body=body)
        writer.write(response, deprecation_message=DEPRECATION_MESSAGE)


@app.command(short_help="Create a new webapp or Update an existing webapp.")
def create_or_update(
    ctx: typer.Context,
    file: Annotated[Optional[str], typer.Argument(..., help=FILE_HELP_STRING)] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    name: str = NAME,
    title: str = TITLE,
    description: str = DESCRIPTION,
    image_id: int = IMAGE_ID,
    version_id: int = VERSION_ID,
    instance_type_id: int = INSTANCE_TYPE_ID,
    session_stickiness: bool = SESSION_STICKINESS,
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new webapp or ***Update*** an existing webapp based on webapp name and start its deployment.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    When updating the webapp, it will trigger a redeployment only under specific conditions.
    Redeployment is triggered if you make changes to any of the following parameters: imageId, versionId, instanceTypeId or sessionStickiness.
    However, only modifying the title or description will not trigger a redeployment.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (string): Name of the webapp.
        title (string | required: false): Title of the webapp.
        description (string | required: false): Description of the webapp.
        imageDetails (map):
            imageId (number): ID of the image.
            versionId (number | required: false): ID of the image version. If versionId is not passed, webapp will be created using latest ready version of the image.
        resources (map | required: false):
            instanceTypeId (number): ID of the instance type.
        sessionStickiness (boolean | required: false): Enable session stickiness for the webapp. Default value is false. Enabling session stickiness will tie each user to a specific server for all their requests.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak webapps create-or-update '/path/to/file.yml' -v '/path/to/params.yml'
    ```

    \b
    📝 ***Example usage without yaml:***
    ```bash
    peak webapps create-or-update --name <name> --title <title> --description <description> --image-id <image-id> --version-id <version-id> --instance-type-id <instance-type-id> --session-stickiness
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/create-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        name,
        description,
        title,
        image_id,
        version_id,
        instance_type_id,
        session_stickiness,
    )
    user_options = map_user_options(user_options, MAPPING)

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, webapp_client.create_or_update_webapp)

    body = combine_dictionaries(body.get("body") or {}, user_options)

    response = webapp_client.create_or_update_webapp(body=body)
    writer.write(response, deprecation_message=DEPRECATION_MESSAGE)


@app.command(short_help="Delete an existing webapp.")
def delete(
    ctx: typer.Context,
    webapp_id: str = _WEBAPP_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** an existing webapp.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak webapps delete "ab11c21d-1add-45dd-a72e-8c6b83b68dee"
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "ab11c21d-1add-45dd-a72e-8c6b83b68dee"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/delete-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = webapp_client.delete_webapp(webapp_id=webapp_id)
        writer.write(response, deprecation_message=DEPRECATION_MESSAGE)


@app.command(short_help="Describe details of a webapp.")
def describe(
    ctx: typer.Context,
    webapp_id: str = _WEBAPP_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** details for the specific webapp.

    \b
     ```yaml
    NOTE: The Web Apps commands are deprecated and will eventually be removed. It is recommended that users begin transitioning to the Service commands, which offers all the features of Webapp along with the added capability of API deployment and other significant updates for enhanced functionality. Transitioning early to the Service commands will ensure easier adoption of its advanced features and future updates.
    ```

    \b
    For more information on Services, please refer to the [**Service**](https://docs.peak.ai/sdk/latest/cli/reference.html#peak-services) documentation.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak webapps describe "ab11c21d-1add-45dd-a72e-8c6b83b68dee"
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": "ab11c21d-1add-45dd-a72e-8c6b83b68dee",
        "name": "awesome-webapp",
        "status": "AVAILABLE",
        "imageDetails": {
            "imageId": 1,
            "versionId": 1
        },
        "resources": {
            "instanceTypeId": 1
        },
        "sessionStickiness": false,
        "createdAt": "2020-01-01T18:00:00.000Z",
        "createdBy": "someone@peak.ai",
        "updatedAt": "2020-01-01T18:00:00.000Z",
        "updatedBy": "someone@peak.ai",
        "tags": [...]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/webapps/api-docs/index.htm#/Services/get-service)
    """
    webapp_client: Webapp = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = webapp_client.describe_webapp(webapp_id=webapp_id)
        writer.write(response, deprecation_message=DEPRECATION_MESSAGE)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Metrics client module."""

from __future__ import annotations

import json
from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ArtifactInfo, ContentType, HttpMethods
from peak.exceptions import InvalidParameterException
from peak.session import Session


class Metric(BaseClient):
    """Client class for interacting with metrics resource."""

    BASE_ENDPOINT = "semantic-layer/api/v1"

    def __get_artifact_details(self, artifact: ArtifactInfo) -> str | None:
        path = None

        if artifact:
            path = artifact.get("path")

        return path

    def publish(
        self,
        artifact: Optional[ArtifactInfo] = None,
        collection_id: Optional[str] = None,
        body: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Publish Metrics.

        The metrics can be published either by passing artifact and namespace,
        or by passing collection_id and namespace. If both artifact and collection_id
        are provided, artifact takes priority. If the namespace is not provided, the 'default' namespace is used.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/post_api_v1_metrics>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the details to publish the metrics.
            artifact (ArtifactInfo | None): Mapping of artifact attributes that specifies how the artifact will be generated,
                                     it accepts one key `path`,
            collection_id (str | None): The ID of the collection to publish the metrics.

        Returns:
            Dict[str, Any]: A dictionary containing details about the published metrics.

        SCHEMA:
            .. code-block:: json

                {
                  "namespace": "string",
                }

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        if artifact:
            method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/metrics"
            path = self.__get_artifact_details(artifact)

            return self.session.create_request(  # type: ignore[no-any-return]
                endpoint,
                method,
                content_type=ContentType.MULTIPART_FORM_DATA,
                body=body,
                path=path,
            )

        if collection_id:
            method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/metrics/collections/{collection_id}/publish"
            return self.session.create_request(  # type: ignore[no-any-return]
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                body=body,
            )

        raise InvalidParameterException(
            message="Either Artifact or Collection ID must be passed to publish the metrics.",
        )

    def query(
        self,
        measures: List[str],
        namespace: Optional[str] = None,
        generate_sql: Optional[bool] = False,  # noqa: FBT002
        dimensions: Optional[List[str]] = None,
        filters: Optional[List[Dict[str, Any]]] = None,
        time_dimensions: Optional[List[Dict[str, Any]]] = None,
        segments: Optional[List[str]] = None,
        order: Optional[Dict[str, Literal["asc", "desc"]]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Query a published metric in the semantic layer using the provided parameters.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Query/get_api_v1_metrics_query>`__

        Args:
            measures (List[str]): An array of measures to include in the query. Measures represent quantitative metrics such as sums or counts.
            namespace (str | None): The namespace associated with the metrics. If not provided, the default namespace is used.
            generate_sql (bool | None): Indicates whether to return the SQL query instead of data. If `true`, the response will include the SQL query used to retrieve the metrics. Default is `false`.
            dimensions (List[str] | None): An array of dimensions to include in the query. Dimensions represent qualitative categories such as time, location, or product names.
            filters (List[Dict[str, Any]] | None): An array of filter objects to apply to the query. Filters limit the data returned based on specific conditions.
                dimension (str): The dimension to filter on.
                operator (str): The operator to use for the filter. Supported values are `equals`, `notEquals`, `contains`, `notContains`, `startsWith`, `notStartsWith`, `endsWith`, `notEndsWith`, `gt`, `gte`, `lt`, `lte`, `inDateRange`, `notInDateRange`, `beforeDate`, `beforeOrOnDate`, `afterDate`, `afterOrOnDate` etc.
                values (List[str]): An array of values to filter on.
            time_dimensions (List[Dict[str, Any]] | None): Time dimensions allow querying over specific time ranges with optional granularity (e.g., day, month, year).
                dimension (str): The time dimension to include in the query.
                granularity (str | None): The granularity of the time dimension. Supported values are `second`, `minute`, `hour`, `day`, `week`, `month`, `quarter`, and `year`.
                dateRange (list(str) | str | None): An array of two dates that define the time range for the query. Alternatively, you can provide a single string out of the following predefined date ranges `today`, `yesterday`, `this week`, `last week`, `this month`, `last month`, `this quarter`, `last quarter`, `this year`, `last year`, `last 7 days` and `last 30 days`.
            segments (List[str] | None): An array of segments to include in the query. Segments represent pre-defined filters that can be applied to metrics.
            order (Dict[str, Any] | None): Defines the sort order of the results. This is an object where keys are the dimensions/measures and values are either 'asc' or 'desc' to specify ascending or descending order.
            limit (int | None): Limits the number of rows returned by the query. If not provided, the default limit is applied.
            offset (int | None): Specifies the number of rows to skip before starting to return data. Useful for pagination.

        Returns:
            Dict[str, Any]: A dictionary containing the query metrics response.

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/metrics/query"

        request_params = {
            "namespace": namespace,
            "generateSql": str(generate_sql).lower(),
            "measures": measures,
            "dimensions": dimensions,
            "filters": json.dumps(filters) if filters else None,
            "timeDimensions": json.dumps(time_dimensions) if time_dimensions else None,
            "segments": segments,
            "order": json.dumps(order) if order else None,
            "limit": limit,
            "offset": offset,
        }

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            params=request_params,
            content_type=ContentType.APPLICATION_JSON,
        )

    @overload
    def list(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        namespace: Optional[str] = None,
        type: Optional[str] = None,  # noqa: A002
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        namespace: Optional[str] = None,
        type: Optional[str] = None,  # noqa: A002
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        namespace: Optional[str] = None,
        type: Optional[str] = None,  # noqa: A002
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the list of metrics.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/get_api_v1_metrics>`__

        Args:
            page_size (int | None): The number of images per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            namespace (str | None): The namespace associated with the metrics. If not provided, the default namespace is used.
            type (str | None): The type of metrics to retrieve. If not provided, all metrics are retrieved. Available types are `cube`, `view`, `dimension`, `measure`, `segment` and `all`.
            return_iterator (bool): Whether to return an iterator object or list of metrics for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/metrics"
        params = {
            "pageSize": page_size,
            "namespace": namespace,
            "type": type,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="data",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def delete(
        self,
        *,
        namespace: Optional[str] = None,
        measures: Optional[List[str]] = None,
        publication_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Delete one or more measures.

        The measures can be deleted either by passing publication_id in which case all metrics related to the publication will be deleted.
        Or by passing namespace and measures in which case only the specified measures will be deleted.
        If both are passed, publication_id takes priority.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Metrics/delete_api_v1_metrics_publications__publicationId_>`__

        Args:
            namespace (str): The namespace to delete the measures from. Required if measures is passed.
            measures (List[str]): An array of measures to delete.
            publication_id (str): The publication ID to delete. Passing this will delete all the metrics in the
                publication.

        Returns:
            Dict[str, Any]: An empty dictionary.

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        if publication_id:
            method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/metrics/publications/{publication_id}"
            return self.session.create_request(  # type: ignore[no-any-return]
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
            )

        if measures:
            method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/metrics"
            body = {"namespace": namespace, "measures": measures}
            return self.session.create_request(  # type: ignore[no-any-return]
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                params=body,
            )

        raise InvalidParameterException(
            message="Either Publication Id or Measures must be passed.",
        )

    def create_collection(
        self,
        artifact: ArtifactInfo,
        body: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Create Metric Collection.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/post_api_v1_metrics_collections>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the details to publish the metrics.
            artifact (ArtifactInfo): Mapping of artifact attributes that specifies how the artifact will be generated,
                                     it accepts one key `path`,

        Returns:
            Dict[str, Any]: A dictionary containing details about the published metrics.

        SCHEMA:
            .. code-block:: json

                {
                  "name": "string",
                  "scope": "string",
                  "description": "string"
                }

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/metrics/collections"

        if not artifact:
            raise InvalidParameterException(
                message="Artifact must be provided to create the metrics collection.",
            )

        path = self.__get_artifact_details(artifact)

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.MULTIPART_FORM_DATA,
            body=body,
            path=path,
        )

    def delete_collection(self, collection_id: str) -> Dict[str, Any]:
        """Delete a metric collection.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/delete_api_v1_metrics_collections__collectionId_>`__

        Args:
            collection_id (str): The ID of the collection to delete.

        Returns:
            Dict[str, Any]: A dictionary containing details about the deleted collection.

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/metrics/collections/{collection_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    @overload
    def list_collections(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        id: Optional[List[str]] = None,  # noqa: A002
        scope: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_collections(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        id: Optional[List[str]] = None,  # noqa: A002
        scope: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_collections(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        id: Optional[List[str]] = None,  # noqa: A002
        scope: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the list of metric collections.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Collections/get_api_v1_metrics_collections>`__

        Args:
            page_size (int | None): The number of images per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            id (List[str] | None): An array of collection IDs to retrieve. If not provided, all collections are retrieved.
            scope (List[str] | None): An array of scopes to filter the collections by. Available scopes are `PUBLIC` and `PRIVATE`. If not provided, all collections of the tenant along with all public collections are retrieved.
            return_iterator (bool): Whether to return an iterator object or list of metrics for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/metrics/collections"
        params = {
            "pageSize": page_size,
            "id": id,
            "scope": scope,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="collections",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    @overload
    def list_namespaces(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_namespaces(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_namespaces(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the list of namespaces.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/semantic-layer/api-docs/index.htm#/Namespaces/get_api_v1_namespaces>`__

        Args:
            page_size (int | None): The number of images per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            return_iterator (bool): Whether to return an iterator object or list of metrics for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/namespaces"
        params = {
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="namespaces",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )


def get_client(session: Optional[Session] = None) -> Metric:
    """Returns a Metrics client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Metric: The metric client object.
    """
    return Metric(session)


__all__ = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Supported Log levels."""
from __future__ import annotations

import logging
from enum import IntEnum
from typing import Final, Literal


class LogLevel(IntEnum):
    """Enumeration of log levels to be used in logging.

    Each enum member corresponds to a specific log level defined in the logging module.
    The enum provides a convenient way to specify log levels when configuring loggers.

    Attributes:
        DEBUG: Debug log level. Intended for detailed debugging information.
        INFO: Info log level. Used for general information about program execution.
        WARN: Warning log level. Indicates potential issues or unexpected behavior.
        WARNING: Warning log level. Indicates potential issues or unexpected behavior.
        ERROR: Error log level. Indicates errors that do not prevent program execution.
        EXCEPTION: Error log level. Indicates errors that do not prevent program execution.
        CRITICAL: Critical log level. Indicates severe errors that might lead to program failure.
        FATAL: Critical log level. Indicates severe errors that might lead to program failure.
    """

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARNING
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    EXCEPTION = logging.ERROR
    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL


LogLevelNames = Literal[
    "DEBUG",
    "INFO",
    "WARN",
    "WARNING",
    "ERROR",
    "EXCEPTION",
    "CRITICAL",
    "FATAL",
]


LOG_LEVEL_NAMES_TO_LOG_LEVEL: Final[dict[LogLevelNames, LogLevel]] = {
    "DEBUG": LogLevel.DEBUG,
    "INFO": LogLevel.INFO,
    "WARNING": LogLevel.WARN,
    "WARN": LogLevel.WARN,
    "ERROR": LogLevel.ERROR,
    "EXCEPTION": LogLevel.ERROR,
    "FATAL": LogLevel.CRITICAL,
    "CRITICAL": LogLevel.CRITICAL,
}
